"""Meltano runtime environments."""

import copy
from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar

from meltano.core.behavior import NameEq
from meltano.core.behavior.canonical import Canonical
from meltano.core.plugin import PluginType
from meltano.core.plugin.base import PluginRef
from meltano.core.setting_definition import SettingDefinition
from meltano.core.utils import NotFound

TEnv = TypeVar("TEnv")


class NoActiveEnvironment(Exception):
    """Exception raised when invocation has no active environment."""


class EnvironmentPluginConfig(PluginRef):
    """Environment configuration for a plugin."""

    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        config: Optional[dict] = None,
        **extras,
    ):
        """Create a new plugin configuration object.

        Args:
            plugin_type: Extractor, loader, etc.
            name: Name of the plugin.
            config: Plugin configuration.
            extras: Plugin extras.
        """
        super().__init__(plugin_type, name)
        self.config = copy.deepcopy(config or {})
        self.extras = extras

    @property
    def extra_config(self):
        """Get extra config from the Meltano environment, like `select` and `schema`.

        Returns:
            Extra config.
        """
        return {f"_{key}": value for key, value in self.extras.items()}

    @property
    def config_with_extras(self):
        """Get plugin configuration values from the Meltano environment.

        Returns:
            Plugin configuration with extra values.
        """
        return {**self.config, **self.extra_config}

    @config_with_extras.setter
    def config_with_extras(self, new_config_with_extras: Dict[str, Any]):
        """Set plugin configuration values from the Meltano environment.

        Args:
            new_config_with_extras: New plugin configuration with extra values.
        """
        self.config.clear()
        self.extras.clear()

        for key, value in new_config_with_extras.items():
            if key.startswith("_"):
                self.extras[key[1:]] = value
            else:
                self.config[key] = value

    def get_orphan_settings(
        self, existing: Iterable[SettingDefinition]
    ) -> List[SettingDefinition]:
        """Get orphan settings for this plugin.

        Orphan settings are `config` entries that do not have a
        matching parent entry within `settings`.

        Args:
            existing: Existing settings.

        Returns:
            List of orphan settings.
        """
        return SettingDefinition.from_missing(existing, self.config)


class EnvironmentConfig(Canonical):
    """Meltano environment configuration."""

    def __init__(self, plugins: Dict[str, List[dict]] = None, **extras):
        """Create a new environment configuration.

        Args:
            plugins: Mapping of plugin types to arrays of plugin configurations.
            extras: Environment extras.
        """
        super().__init__(extras=extras)
        self.plugins = self.load_plugins(plugins or {})

    def load_plugins(
        self,
        plugins: Dict[str, List[dict]],
    ) -> Dict[PluginType, List[EnvironmentPluginConfig]]:
        """Create plugin configurations from raw dictionary.

        Args:
            plugins: Plugin configurations.

        Returns:
            A mapping of plugin type to arrays of plugin configurations.
        """
        plugin_mapping = {}
        for raw_plugin_type, raw_plugins in plugins.items():
            plugin_type = PluginType(raw_plugin_type)
            plugin_mapping[plugin_type] = [
                EnvironmentPluginConfig(plugin_type=plugin_type, **raw_plugin)
                for raw_plugin in raw_plugins
            ]

        return plugin_mapping


class Environment(NameEq, Canonical):
    """Runtime environment for Meltano runs."""

    def __init__(self, name: str, config: dict = None, env: dict = None) -> None:
        """Create a new environment object.

        Args:
            name: Environment name. Must be unique. Defaults to None.
            config: Dictionary with environment configuration.
            env: Optional override environment values.
        """
        super().__init__()

        self.name = name
        self.config = EnvironmentConfig(**(config or {}))
        self.env = env or {}

    @classmethod
    def find(cls: Type[TEnv], objects: Iterable[TEnv], name: str) -> TEnv:
        """Lookup an environment by name from an iterable.

        Args:
            objects: Iterable of objects to search.
            name: Environment name.

        Returns:
            Environment object if found.

        Raises:
            NotFound: If an environment is not found.
        """
        try:
            return next(env for env in objects if env.name == name)
        except StopIteration as stop:
            raise NotFound(name, cls) from stop

    def get_plugin_config(
        self,
        plugin_type: PluginType,
        name: str,
    ) -> EnvironmentPluginConfig:
        """Get configuration for a plugin in this environment.

        Args:
            plugin_type: Extractor, loader, etc.
            name: Plugin name.

        Returns:
            A plugin configuration object if one is present in this environment.
        """
        default = EnvironmentPluginConfig(plugin_type, name)

        return next(
            filter(
                lambda plugin: plugin.name == name,
                self.config.plugins.get(plugin_type, []),
            ),
            default,
        )



"""Legacy tracker for Meltano.

This module and its content will be removed in a future version of Meltano without a major version
bump. It only remains here to make the transition to the new tracker as smooth as possible.
"""

from __future__ import annotations

from typing import Any

from snowplow_tracker import SelfDescribingJson

from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.schedule import Schedule
from meltano.core.tracking import ProjectContext, Tracker
from meltano.core.utils import hash_sha256

REQUEST_TIMEOUT = 2.0
MEASUREMENT_PROTOCOL_URI = "https://www.google-analytics.com/collect"
DEBUG_MEASUREMENT_PROTOCOL_URI = "https://www.google-analytics.com/debug/collect"


class LegacyTracker:  # noqa: WPS214, WPS230
    """Legacy tracker for Meltano."""

    def __init__(
        self,
        project: Project,
        tracking_id: str = None,
        request_timeout: float = None,
        context_overrides: tuple[SelfDescribingJson] = None,
    ):
        """Create a new Google Analytics tracker.

        Args:
            project: Meltano project.
            tracking_id: Unique identifier for tracking. Defaults to None.
            request_timeout: For GA requests. Defaults to None.
            context_overrides: A list of explicit context overrides that will be set on the underlying Tracker.
        """
        self.project = project
        self.settings_service = ProjectSettingsService(self.project)

        self.tracking_id = tracking_id or self.settings_service.get("tracking_ids.cli")
        self.request_timeout = request_timeout or REQUEST_TIMEOUT

        self.send_anonymous_usage_stats = self.settings_service.get(
            "send_anonymous_usage_stats", True
        )

        self.tracker = Tracker(
            project,
            request_timeout=self.request_timeout,
        )
        if context_overrides:
            self.tracker._contexts = context_overrides

        project_context = ProjectContext(project, self.tracker.client_id)
        self.project_id = project_context.project_uuid

    def event(self, category: str, action: str) -> dict[str, Any]:
        """Construct a GA event with all the required parameters.

        Args:
            category: The category of the event.
            action: The action of the event.

        Returns:
            A dict with all the required parameters.
        """
        return {
            "v": "1",
            "tid": self.tracking_id,
            "cid": self.tracker.client_id,
            "ds": "meltano cli",
            "t": "event",
            "ec": category,
            "ea": action,
            "el": self.project_id,
            "cd1": self.project_id,  # maps to the custom dimension 1 of the UI
        }

    def track_event(self, category: str, action: str, debug: bool = False) -> None:
        """Send a struct event to Snowplow.

        Args:
            category: GA event category.
            action: GA event action.
            debug: If True, send the event to the debug endpoint.
        """
        self.tracker.track_struct_event(category, action)

    def track_meltano_init(self, project_name: str, debug: bool = False) -> None:
        """Track the initialization of a Meltano project.

        Args:
            project_name: The name of the project.
            debug: Whether to send the event to the debug endpoint.
        """
        self.track_event(
            category="meltano init", action=f"meltano init {project_name}", debug=debug
        )

    def track_meltano_add(
        self, plugin_type: str, plugin_name: str, debug: bool = False
    ) -> None:
        """Track a plugin add event.

        Args:
            plugin_type: The type of the plugin.
            plugin_name: The name of the plugin.
            debug: Whether to send the event to the debug endpoint.
        """
        self.track_event(
            category=f"meltano add {plugin_type}",
            action=f"meltano add {plugin_type} {plugin_name}",
            debug=debug,
        )

    def track_meltano_discover(self, plugin_type: str, debug: bool = False) -> None:
        """Track the discovery of a plugin type.

        Args:
            plugin_type: The type of plugin that was discovered.
            debug: Whether to send the event to the debug endpoint.
        """
        self.track_event(
            category="meltano discover",
            action=f"meltano discover {plugin_type}",
            debug=debug,
        )

    def track_meltano_elt(
        self, extractor: str, loader: str, transform: str, debug: bool = False
    ) -> None:
        """Track a meltano elt command.

        Args:
            extractor: The extractor name.
            loader: The loader name.
            transform: The transform name.
            debug: Whether to send the event to the debug endpoint.
        """
        self.track_event(
            category="meltano elt",
            action=f"meltano elt {extractor} {loader} --transform {transform}",
            debug=debug,
        )

    def track_meltano_install(self, debug: bool = False) -> None:
        """Track the installation of plugins.

        Args:
            debug: Whether to send the event to the debug endpoint.
        """
        self.track_event(
            category="meltano install", action="meltano install", debug=debug
        )

    def track_meltano_invoke(
        self, plugin_name: str, plugin_args: str, debug: bool = False
    ) -> None:
        """Track a meltano invoke event.

        Args:
            plugin_name: The name of the plugin invoked.
            plugin_args: The arguments passed to the plugin.
            debug: Whether to send the event to the debug endpoint.
        """
        self.track_event(
            category="meltano invoke",
            action=f"meltano invoke {plugin_name} {plugin_args}",
            debug=debug,
        )

    def track_meltano_schedule(
        self, action: str, schedule: Schedule | None = None, debug: bool = False
    ) -> None:
        """Track a schedule event.

        Args:
            action: The type of action taken on the schedule (e.g. add, run, list)
            schedule: The schedule to track.
            debug: Whether to send the event to the debug endpoint.
        """
        if schedule:
            if schedule.job:
                self.track_event(
                    category="meltano schedule",
                    action=(
                        f"meltano schedule {action} {schedule.name} "
                        + f"--job={schedule.job} --interval={schedule.interval}"
                    ),
                    debug=debug,
                )
            else:
                self.track_event(
                    category="meltano schedule",
                    action=(
                        f"meltano schedule {action} {schedule.name} "
                        + f"--extractor {schedule.extractor} --loader {schedule.loader} --interval {schedule.interval} "
                        + f"--transform={schedule.transform}"
                    ),
                    debug=debug,
                )
        else:
            self.track_event(
                category="meltano schedule",
                action=(f"meltano schedule {action}"),
                debug=debug,
            )

    def track_meltano_select(
        self,
        extractor: str,
        entities_filter: str,
        attributes_filter: str,
        flags: dict[str, bool],
        debug: bool = False,
    ) -> None:
        """Track the selection of entities and attributes.

        Args:
            extractor: The extractor name.
            entities_filter: The entities filter.
            attributes_filter: The attributes filter.
            flags: The CLI flags.
            debug: Whether to send the event to the debug endpoint.
        """
        action = f"meltano select {extractor} {entities_filter} {attributes_filter}"

        if flags["list"]:
            action = f"{action} --list"
        if flags["all"]:
            action = f"{action} --all"
        if flags["exclude"]:
            action = f"{action} --exclude"

        self.track_event(category="meltano select", action=action, debug=debug)

    def track_meltano_state(self, subcommand: str, state_id: str | None = None):
        """Track the management of Singer State.

        Args:
            subcommand: The subcommand being run (e.g. 'set' or 'clear')
            state_id: The state_id for which state is being managed
        """
        action = f"meltano state {subcommand}"
        if state_id:
            hashed_state_id = hash_sha256(state_id)
            action = f"{action} {hashed_state_id}"
        self.track_event(category="meltano state", action=action)

    def track_meltano_ui(self, debug: bool = False) -> None:
        """Track the UI start-up.

        Args:
            debug: Whether to send the event to the debug endpoint.
        """
        action = "meltano ui"
        self.track_event(category="meltano ui", action=action, debug=debug)

    def track_meltano_test(
        self,
        plugin_tests: tuple[str],
        debug: bool = False,
        **flags: dict[str, bool],
    ) -> None:
        """Track invocations of `meltano test`.

        Args:
            plugin_tests: A tuple of plugin names and test names.
            debug: Whether to send the event to the debug endpoint.
            flags: A dictionary of CLI flags.
        """
        action = "meltano test"

        if flags["all_tests"]:
            action = f"{action} --all"

        for plugin_test in plugin_tests:
            action = f"{action} {plugin_test}"

        self.track_event(
            category="meltano test",
            action=action,
            debug=debug,
        )

    def track_meltano_run(self, blocks: list[str], debug: bool = False) -> None:
        """Track invocations of `meltano run`.

        Args:
            blocks: The blocks to track.
            debug: Whether to print debug information.
        """
        blocks_string = " ".join(blocks)
        self.track_event(
            category="meltano run",
            action=f"meltano run {blocks_string}",
            debug=debug,
        )

    def track_meltano_job(
        self, action: str, job_name: str | None = None, debug: bool = False
    ) -> None:
        """Track a job event.

        Args:
            action: The type of action taken on the job (e.g. add, run, list)
            job_name: The job to track.
            debug: Whether to send the event to the debug endpoint.
        """
        if job_name:
            self.track_event(
                category="meltano job",
                action=(f"meltano job {job_name} {action} "),
                debug=debug,
            )
        else:
            self.track_event(
                category="meltano job",
                action=f"meltano job {action}",
                debug=debug,
            )



"""New Project Initialization CLI."""
import logging

import click

from meltano.core.error import SubprocessError
from meltano.core.legacy_tracking import LegacyTracker
from meltano.core.project_init_service import ProjectInitService
from meltano.core.project_settings_service import ProjectSettingsService

from . import cli
from .params import database_uri_option
from .utils import CliError

EXTRACTORS = "extractors"
LOADERS = "loaders"
ALL = "all"

logger = logging.getLogger(__name__)


@cli.command(short_help="Create a new Meltano project.")
@click.pass_context
@click.argument("project_name", required=False)
@click.option(
    "--no_usage_stats", help="Do not send anonymous usage stats.", is_flag=True
)
@database_uri_option
def init(ctx, project_name, no_usage_stats):
    """
    Create a new Meltano project.

    Read more at https://docs.meltano.com/reference/command-line-interface#init

    """
    if not project_name:
        click.echo("We need a project name to get started!")
        project_name = click.prompt("Enter a name now to create a Meltano project")

    if ctx.obj["project"]:
        root = ctx.obj["project"].root
        logging.warning(f"Found meltano project at: {root}")
        raise CliError("`meltano init` cannot run inside a Meltano project.")

    if no_usage_stats:
        ProjectSettingsService.config_override["send_anonymous_usage_stats"] = False

    init_service = ProjectInitService(project_name)
    try:  # noqa: WPS229
        project = init_service.init()
        init_service.echo_instructions()
        tracker = LegacyTracker(project)
        tracker.track_meltano_init(project_name=project_name)
    except SubprocessError as err:
        logger.error(err.stderr)
        raise



"""Meltano telemetry contexts for the CLI events."""
from __future__ import annotations

from enum import Enum, auto

from snowplow_tracker import SelfDescribingJson

from meltano.core.tracking.schemas import CliContextSchema


class CliEvent(Enum):
    """The kind of event that is occuring in the command-line interface."""

    started = auto()
    completed = auto()
    skipped = auto()
    failed = auto()
    aborted = auto()


class CliContext(SelfDescribingJson):
    """CLI context for the Snowplow tracker."""

    def __init__(
        self,
        command: str,
        sub_command: str | None = None,
        option_keys: list(str) | None = None,
    ):
        """Initialize a CLI context.

        Args:
            command: The command name e.g. `schedule`.
            sub_command: The sub-command name e.g. `add` or `set`.
            option_keys: The list of option keys e.g. `loader`, `job`.
        """
        super().__init__(
            CliContextSchema.url,
            {
                "command": command,
                "sub_command": sub_command,
                "option_keys": option_keys or [],
            },
        )

    @classmethod
    def from_command_and_kwargs(
        cls, command: str, sub_command: str | None = None, **kwargs
    ) -> CliContext:
        """Initialize a CLI context.

        Args:
            command: The CLI command.
            sub_command: The CLI sub command.
            kwargs: Additional key-value pairs to evaluate as option_keys if they are not None/False.

        Returns:
            A CLI context.
        """
        return cls(
            command=command,
            sub_command=sub_command,
            option_keys=[key for key, val in kwargs.items() if val],
        )



import logging
import os
from contextlib import contextmanager

from meltano.core.project import Project
from meltano.core.utils import makedirs, slugify

MAX_FILE_SIZE = 2_097_152  # 2MB max


class MissingJobLogException(Exception):
    """Occurs when JobLoggingService can not find a requested log."""


class SizeThresholdJobLogException(Exception):
    """Occurs when a Job log exceeds the MAX_FILE_SIZE."""


class JobLoggingService:
    def __init__(self, project: Project):
        self.project = project

    @makedirs
    def logs_dir(self, state_id, *joinpaths):
        return self.project.job_logs_dir(state_id, *joinpaths)

    def generate_log_name(
        self, state_id: str, run_id: str, file_name: str = "elt.log"
    ) -> str:
        """Generate an internal etl log path and name."""
        return self.logs_dir(state_id, str(run_id), file_name)

    @contextmanager
    def create_log(self, state_id, run_id, file_name="elt.log"):
        """
        Open a new log file for logging and yield it.

        Log will be created inside the logs_dir, which is .meltano/logs/elt/:state_id/:run_id
        """
        log_file_name = self.generate_log_name(state_id, run_id, file_name)

        try:
            log_file = open(log_file_name, "w")
        except OSError:
            # Don't stop the Job running if you can not open the log file
            # for writting: just return /dev/null
            logging.error(
                f"Could open log file {log_file_name} for writting. Using /dev/null"
            )
            log_file = open(os.devnull, "w")

        try:
            yield log_file
        finally:
            log_file.close()

    def get_latest_log(self, state_id):
        """
        Get the contents of the most recent log for any ELT job
         that ran with the provided state_id
        """
        try:
            latest_log = next(iter(self.get_all_logs(state_id)))

            if latest_log.stat().st_size > MAX_FILE_SIZE:
                raise SizeThresholdJobLogException(
                    f"The log file size exceeds '{MAX_FILE_SIZE}'"
                )

            with latest_log.open() as f:
                return f.read()
        except StopIteration:
            raise MissingJobLogException(
                f"Could not find any log for job with id '{state_id}'"
            )
        except FileNotFoundError:
            raise MissingJobLogException(
                f"Cannot log for job with id '{state_id}': '{latest_log}' is missing."
            )

    def get_downloadable_log(self, state_id):
        """
        Get the `*.log` file of the most recent log for any ELT job
         that ran with the provided state_id
        """
        try:
            latest_log = next(iter(self.get_all_logs(state_id)))
            return str(latest_log.resolve())
        except StopIteration:
            raise MissingJobLogException(
                f"Could not find any log for job with id '{state_id}'"
            )
        except FileNotFoundError:
            raise MissingJobLogException(
                f"Cannot log for job with id '{state_id}': '{latest_log}' is missing."
            )

    def get_all_logs(self, state_id):
        """
        Get all the log files for any ELT job that ran with the provided state_id

        The result is ordered so that the most recent is first on the list
        """
        log_files = []
        for logs_dir in self.logs_dirs(state_id):
            log_files.extend(list(logs_dir.glob("**/*.log")))

        log_files.sort(key=lambda path: os.stat(path).st_ctime_ns, reverse=True)

        return log_files

    def delete_all_logs(self, state_id):
        """
        Delete all the log files for any ELT job that ran with the provided state_id
        """

        for log_path in self.get_all_logs(state_id):
            log_path.unlink()

    def legacy_logs_dir(self, state_id, *joinpaths):
        job_dir = self.project.run_dir("elt").joinpath(slugify(state_id), *joinpaths)
        return job_dir if job_dir.exists() else None

    def logs_dirs(self, state_id, *joinpaths):
        logs_dir = self.logs_dir(state_id, *joinpaths)
        legacy_logs_dir = self.legacy_logs_dir(state_id, *joinpaths)

        dirs = [logs_dir]
        if legacy_logs_dir:
            dirs.append(legacy_logs_dir)

        return dirs



import click

from meltano.core.db import DB, project_engine

from . import cli
from .params import pass_project


@cli.group(short_help="Manage system DB schema.")
def schema():
    """Manage system DB schema."""


@schema.command(short_help="Create system DB schema, if not exists.")
@click.argument("schema")
@click.argument("roles", nargs=-1, required=True)
@pass_project()
def create(project, schema, roles):
    """Create system DB schema, if not exists."""
    engine, _ = project_engine(project)
    DB.ensure_schema_exists(engine, schema, grant_roles=roles)



"""Defines Job model class."""
import asyncio
import os
import signal
import uuid
from contextlib import contextmanager, suppress
from datetime import datetime, timedelta
from enum import Enum

from async_generator import asynccontextmanager
from sqlalchemy import Column, literal, types
from sqlalchemy.ext.hybrid import Comparator, hybrid_property
from sqlalchemy.ext.mutable import MutableDict

from meltano.core.error import Error
from meltano.core.models import SystemModel
from meltano.core.sqlalchemy import GUID, IntFlag, JSONEncodedDict

HEARTBEATLESS_JOB_VALID_HOURS = 24
HEARTBEAT_VALID_MINUTES = 5


class InconsistentStateError(Error):
    """Occur upon a wrong operation for the current state."""


class ImpossibleTransitionError(Error):
    """Occur upon a wrong transition."""


class State(Enum):
    """Represents status of a Job."""

    IDLE = (0, ("RUNNING", "FAIL"))
    RUNNING = (1, ("SUCCESS", "FAIL"))
    SUCCESS = (2, ())
    FAIL = (3, ("RUNNING",))
    DEAD = (4, ())
    STATE_EDIT = (5, ())

    def transitions(self):
        """Get possible next States for a job of this State.

        Returns:
            The possible states jobs in this state can be transitioned into
        """
        return self.value[1]

    def __str__(self):
        """Get a string representation of this State.

        Returns:
            the name of this State
        """
        return self.name


class StateComparator(Comparator):
    """Compare Job._state to State enums."""

    def __eq__(self, other):
        """Enable SQLAlchemy to directly compare Job.state values with State.

        Args:
            other: the State enum to compare to

        Returns:
            Comparison between __clause_element__ and SQLAlchemy literal for State name
        """
        return self.__clause_element__() == literal(other.name)


def current_trigger():
    """Get the trigger for running job.

    Returns:
        The trigger for currently running job
    """
    return os.getenv("MELTANO_JOB_TRIGGER")


class Payload(IntFlag):
    """Flag indicating whether a Job has state in its payload field."""

    STATE = 1
    INCOMPLETE_STATE = 2


class Job(SystemModel):  # noqa: WPS214
    """Model class that represents a `meltano elt` run in the system database.

    Includes State.STATE_EDIT rows which represent CLI invocations of the
    `meltano state` command which wrote state to the db. Queries that are
    meant to return only actual job runs should filter out records with
    state == State.STATE_EDIT.
    """

    __tablename__ = "job"

    id = Column(types.Integer, primary_key=True)
    job_id = Column(types.String)
    run_id = Column(GUID, nullable=False, default=uuid.uuid4)
    _state = Column(name="state", type_=types.String)
    started_at = Column(types.DateTime)
    last_heartbeat_at = Column(types.DateTime)
    ended_at = Column(types.DateTime)
    payload = Column(MutableDict.as_mutable(JSONEncodedDict))
    payload_flags = Column(IntFlag, default=0)
    trigger = Column(types.String, default=current_trigger)

    def __init__(self, **kwargs):
        """Construct a Job.

        Args:
            kwargs: keyword args to override defaults and pass to super
        """
        kwargs["_state"] = kwargs.pop("state", State.IDLE).name
        kwargs["payload"] = kwargs.get("payload", {})
        kwargs["run_id"] = kwargs.get("run_id", uuid.uuid4())
        super().__init__(**kwargs)

    @hybrid_property
    def state(self) -> State:  # noqa: WPS440
        """Get the job state as a State enum.

        Returns:
            State enum matching string value for this job state
        """
        return State[self._state]

    @state.setter
    def state(self, value):  # noqa: WPS440
        """Set the _state value for this Job from a State enum.

        Args:
            value: the State enum to use.
        """
        self._state = str(value)

    @state.comparator  # noqa: WPS440
    def state(cls):  # noqa: N805, WPS440
        """Use this comparison to compare Job.state to State.

        See:
            https://docs.sqlalchemy.org/en/14/orm/extensions/hybrid.html#building-custom-comparators

        Returns:
            Result of comparison
        """
        return StateComparator(cls._state)

    def is_running(self):
        """Return whether Job is running.

        Returns:
            bool indicating whether this Job is running
        """
        return self.state is State.RUNNING

    def is_stale(self):
        """Return whether Job has gone stale.

        Running jobs with a heartbeat are considered stale after no heartbeat is recorded for 5 minutes.
        Legacy jobs without a heartbeat are considered stale after being in the running state for 24 hours.

        Returns:
            bool indicating whether this Job is stale
        """
        if not self.is_running():
            return False

        if self.last_heartbeat_at:
            timestamp = self.last_heartbeat_at
            valid_for = timedelta(minutes=HEARTBEAT_VALID_MINUTES)
        else:
            timestamp = self.started_at
            valid_for = timedelta(hours=HEARTBEATLESS_JOB_VALID_HOURS)

        return datetime.utcnow() - timestamp > valid_for

    def has_error(self):
        """Return whether a job has failed.

        Returns:
            bool indicating whether this Job has failed
        """
        return self.state is State.FAIL

    def is_complete(self):
        """Return whether a job has completed.

        Returns:
            bool indicating whether this job has completed
        """
        return self.state in {State.SUCCESS, State.FAIL}

    def is_success(self):
        """Return whether a job has succeeded.

        Returns:
            a bool indicating whether this job has succeeded
        """
        return self.state is State.SUCCESS

    def can_transit(self, state: State) -> bool:
        """Return whether this job can transit into the given state.

        Args:
            state: the state to check against

        Returns:
            bool indicating whether the given state is transitable from this job's state
        """
        if self.state is state:
            return True

        return state.name in self.state.transitions()

    def transit(self, state: State) -> (State, State):
        """Transition this job into the given state.

        Args:
            state: the state to transition this job to

        Returns:
            a tuple with the original state and the new state

        Raises:
            ImpossibleTransitionError: when this job cannot transition into the given state
        """
        transition = (self.state, state)

        if not self.can_transit(state):
            raise ImpossibleTransitionError(transition)

        if self.state is state:
            return transition

        self.state = state

        return transition

    @asynccontextmanager
    async def run(self, session):
        """Run wrapped code in context of a job.

        Transitions state to RUNNING and SUCCESS/FAIL as appropriate and records heartbeat every second.

        Args:
            session: the session to use for writing to the db

        Raises:
            BaseException: re-raises an exception occurring in the job running in this context
        """  # noqa: DAR301
        try:
            self.start()
            self.save(session)

            with self._handling_sigterm(session):
                async with self._heartbeating(session):
                    yield

            self.success()
            self.save(session)
        except BaseException as err:  # noqa: WPS424
            if not self.is_running():
                raise

            self.fail(error=self._error_message(err))
            self.save(session)

            raise

    def start(self):
        """Mark the job has having started."""
        self.started_at = datetime.utcnow()
        self.transit(State.RUNNING)

    def fail(self, error=None):
        """Mark the job as having failed.

        Args:
            error: the error to associate with the job's failure
        """
        self.ended_at = datetime.utcnow()
        self.transit(State.FAIL)
        if error:
            self.payload.update({"error": str(error)})

    def success(self):
        """Mark the job as having succeeded."""
        self.ended_at = datetime.utcnow()
        self.transit(State.SUCCESS)

    def fail_stale(self):
        """Mark job as failed if it's gone stale.

        Returns:
            False if job is not stale, else True
        """
        if not self.is_stale():
            return False

        if self.last_heartbeat_at:
            reason = f"No heartbeat recorded for {HEARTBEAT_VALID_MINUTES} minutes."
        else:
            reason = f"Still running after {HEARTBEATLESS_JOB_VALID_HOURS} hours."

        self.fail(f"{reason} The process was likely killed unceremoniously.")

        return True

    def __repr__(self):
        """Represent as a string.

        Returns:
            a string representation of the job
        """
        return f"<Job(id='{self.id}', job_id='{self.job_id}', state='{self.state}', started_at='{self.started_at}', ended_at='{self.ended_at}')>"

    def save(self, session):
        """Save the job in the db.

        Args:
            session: the session to use in querying the db

        Returns:
            the saved job
        """
        session.add(self)
        session.commit()

        return self

    def _heartbeat(self):
        """Update last_heartbeat_at for this job in the db."""
        self.last_heartbeat_at = datetime.utcnow()

    async def _heartbeater(self, session):
        """Heartbeat to the db every second.

        Args:
            session: the session to use for writing to the db
        """
        while True:  # noqa: WPS457
            self._heartbeat()
            self.save(session)

            await asyncio.sleep(1)

    @asynccontextmanager
    async def _heartbeating(self, session):
        """Provide a context for heartbeating jobs.

        Args:
            session: the session to use for writing to the db
        """  # noqa: DAR301
        heartbeat_future = asyncio.ensure_future(self._heartbeater(session))
        try:
            yield
        finally:
            if not heartbeat_future.cancelled():
                heartbeat_future.cancel()

            with suppress(asyncio.CancelledError):
                await heartbeat_future

    @contextmanager
    def _handling_sigterm(self, session):
        def handler(*_):  # noqa: WPS430
            sigterm_status = 143
            raise SystemExit(sigterm_status)

        original_termination_handler = signal.signal(signal.SIGTERM, handler)

        try:
            yield
        finally:
            signal.signal(signal.SIGTERM, original_termination_handler)

    def _error_message(self, err):
        if isinstance(err, SystemExit):
            return "The process was terminated"

        if isinstance(err, (KeyboardInterrupt, asyncio.CancelledError)):
            return "The process was interrupted"

        if str(err):
            return str(err)

        return repr(err)



"""Module for handling multiple project .yml files."""

import copy
import logging
from collections import OrderedDict
from pathlib import Path
from typing import List

import yaml
from atomicwrites import atomic_write

logger = logging.getLogger(__name__)

BLANK_SUBFILE = {"plugins": {}, "schedules": []}  # noqa: WPS407


def deep_merge(parent: dict, children: List[dict]) -> dict:
    """Deep merge a list of child dicts with a given parent."""
    base = copy.deepcopy(parent)
    for child in children:
        for key, value in child.items():
            if isinstance(value, dict):
                # get node or create one
                node = base.setdefault(key, {})
                base[key] = deep_merge(node, [value])
            elif isinstance(value, list):
                node = base.setdefault(key, [])
                node.extend(value)
            else:
                base[key] = value
    return base


class InvalidIncludePath(Exception):
    """Occurs when an included file path matches a provided pattern but is not a valid config file."""


class ProjectFiles:  # noqa: WPS214
    """Interface for working with multiple project yaml files."""

    def __init__(self, root: Path, meltano_file_path: Path):
        """Instantiate ProjectFiles interface from project root and meltano.yml path."""
        self.root = root.resolve()
        self._meltano = None
        self._meltano_file_path = meltano_file_path.resolve()
        self._plugin_file_map = {}

    @property
    def meltano(self):
        """Return the contents of this projects `meltano.yml`."""
        if self._meltano is None:
            with open(self._meltano_file_path) as melt_f:
                self._meltano = yaml.safe_load(melt_f)
        return self._meltano

    @property
    def include_paths(self) -> List[Path]:
        """Return list of paths derived from glob patterns defined in the meltanofile."""
        include_path_patterns = self.meltano.get("include_paths", [])
        return self._resolve_include_paths(include_path_patterns)

    def load(self) -> dict:
        """Load all project files into a single dict representation."""
        # meltano file may have changed in another process, so reset cache first
        self.reset_cache()
        included_file_contents = self._load_included_files()
        return deep_merge(self.meltano, included_file_contents)

    def update(self, meltano_config: dict):
        """Update config by overriding current config with new, changed config.

        Note: `.update()` will write blank entities for those no longer in use (i.e. contained config on load, but not on save).
        """
        file_dicts = self._split_config_dict(meltano_config)
        for file_path, contents in file_dicts.items():
            self._write_file(file_path, contents)

        unused_files = [fl for fl in self.include_paths if str(fl) not in file_dicts]
        for unused_file_path in unused_files:
            self._write_file(unused_file_path, BLANK_SUBFILE)
        self.reset_cache()
        return meltano_config

    def reset_cache(self):
        """Reset cached view of the meltano.yml file."""
        self._meltano = None

    def _is_valid_include_path(self, file_path: Path):
        """Determine if given path is a valid `include_paths` file."""
        if not (file_path.is_file() and file_path.exists()):
            raise InvalidIncludePath(f"Included path '{file_path}' not found.")

    def _resolve_include_paths(self, include_path_patterns: List[str]) -> List[Path]:
        """Return a list of paths from a list of glob pattern strings, not including `meltano.yml` (even if it is matched by a pattern)."""
        include_paths = []
        for pattern in include_path_patterns:
            for path in self.root.glob(pattern):
                try:
                    self._is_valid_include_path(path)
                except InvalidIncludePath as err:
                    logger.critical(f"Include path '{path}' is invalid: \n {err}")
                    raise err
                include_paths.append(path)
            if self._meltano_file_path in include_paths:
                include_paths.remove(self._meltano_file_path)

        # Deduplicate entries
        return list(OrderedDict.fromkeys(include_paths))

    def _add_to_index(self, key: tuple, include_path: Path) -> None:
        """Add a new key:path to the `_plugin_file_map`."""
        if key in self._plugin_file_map:
            key_path_string = ":".join(key)
            existing_key_file_path = self._plugin_file_map.get(key)
            logger.critical(
                f'Plugin with path "{key_path_string}" already added in file {existing_key_file_path}.'
            )
            raise Exception("Duplicate plugin name found.")
        else:
            self._plugin_file_map.update({key: str(include_path)})

    def _index_file(  # noqa: WPS210
        self, include_file_path: Path, include_file_contents: dict
    ) -> None:
        """Populate map of plugins/schedules to their respective files.

        This allows us to know exactly which plugin is configured where when we come to update plugins.
        """
        # index plugins
        all_plugins = include_file_contents.get("plugins", {})
        for plugin_type, plugins in all_plugins.items():
            for plugin in plugins:
                plugin_key = ("plugins", plugin_type, plugin["name"])
                self._add_to_index(key=plugin_key, include_path=include_file_path)
        # index schedules
        schedules = include_file_contents.get("schedules", [])
        for schedule in schedules:
            schedule_key = ("schedules", schedule["name"])
            self._add_to_index(key=schedule_key, include_path=include_file_path)
        # index environments
        environments = include_file_contents.get("environments", [])
        for environment in environments:
            environment_key = ("environments", environment["name"])
            self._add_to_index(key=environment_key, include_path=include_file_path)

    def _load_included_files(self):
        """Read and index included files."""
        self._plugin_file_map = {}
        included_file_contents = []
        for path in self.include_paths:
            try:
                with path.open() as file:
                    contents = yaml.safe_load(file)
                    # TODO: validate dict schema (https://gitlab.com/meltano/meltano/-/issues/3029)
                    self._index_file(
                        include_file_path=path, include_file_contents=contents
                    )
                    included_file_contents.append(contents)
            except yaml.YAMLError as exc:
                logger.critical(f"Error while parsing YAML file: {path} \n {exc}")
                raise exc
        return included_file_contents

    @staticmethod
    def _add_plugin(file_dicts, file, plugin_type, plugin):
        file_dict = file_dicts.setdefault(file, {})
        plugins_dict = file_dict.setdefault("plugins", {})
        plugins = plugins_dict.setdefault(plugin_type, [])
        if plugin["name"] not in {plg["name"] for plg in plugins}:
            plugins.append(plugin)

    @staticmethod
    def _add_schedule(file_dicts, file, schedule):
        file_dict = file_dicts.setdefault(file, {})
        schedules = file_dict.setdefault("schedules", [])
        if schedule["name"] not in {scd["name"] for scd in schedules}:
            schedules.append(schedule)

    @staticmethod
    def _add_environment(file_dicts, file, environment):
        file_dict = file_dicts.setdefault(file, {})
        environments = file_dict.setdefault("environments", [])
        if environment["name"] not in {env["name"] for env in environments}:
            environments.append(environment)

    def _add_plugins(self, file_dicts, all_plugins):
        for plugin_type, plugins in all_plugins.items():
            plugin_type = str(plugin_type)
            for plugin in plugins:
                key = ("plugins", plugin_type, plugin.get("name"))
                file = self._plugin_file_map.get(key, str(self._meltano_file_path))
                self._add_plugin(file_dicts, file, plugin_type, plugin)

    def _add_schedules(self, file_dicts, schedules):
        for schedule in schedules:
            key = ("schedules", schedule["name"])
            file = self._plugin_file_map.get(key, str(self._meltano_file_path))
            self._add_schedule(file_dicts, file, schedule)

    def _add_environments(self, file_dicts, environments):
        for environment in environments:
            key = ("environments", environment["name"])
            file = self._plugin_file_map.get(key, str(self._meltano_file_path))
            self._add_environment(file_dicts, file, environment)

    def _split_config_dict(self, config: dict):
        file_dicts = {}
        for key, value in config.items():
            if key == "plugins":
                self._add_plugins(file_dicts, value)
            elif key == "schedules":
                self._add_schedules(file_dicts, value)
            elif key == "environments":
                self._add_environments(file_dicts, value)
            else:
                file = str(self._meltano_file_path)
                file_dict = file_dicts.setdefault(file, {})
                file_dict[key] = value
        return file_dicts

    @staticmethod
    def _write_file(file_path, contents):
        with atomic_write(file_path, overwrite=True) as fl:
            yaml.dump(contents, fl, default_flow_style=False, sort_keys=False)



import logging  # noqa: D100
import sys
import warnings  # noqa: F401

import click

import meltano
from meltano.core.behavior.versioned import IncompatibleVersionError
from meltano.core.logging import LEVELS, setup_logging
from meltano.core.project import Project, ProjectNotFound
from meltano.core.project_settings_service import ProjectSettingsService

logger = logging.getLogger(__name__)


@click.group(invoke_without_command=True, no_args_is_help=True)
@click.option("--log-level", type=click.Choice(LEVELS.keys()))
@click.option(
    "--log-config", type=str, help="Path to a python logging yaml config file."
)
@click.option("-v", "--verbose", count=True, help="Not used.")
@click.option(
    "--environment",
    envvar="MELTANO_ENVIRONMENT",
    help="Meltano environment name.",
)
@click.option(
    "--no-environment", is_flag=True, default=False, help="Don't use any environment."
)
@click.version_option(version=meltano.__version__, prog_name="meltano")
@click.pass_context
def cli(  # noqa: WPS231
    ctx,
    log_level: str,
    log_config: str,
    verbose: int,
    environment: str,
    no_environment: bool,
):  # noqa: WPS231
    """
    ELT for the DataOps era.

    \b\nRead more at https://www.meltano.com/docs/command-line-interface.html
    """
    if log_level:
        ProjectSettingsService.config_override["cli.log_level"] = log_level

    if log_config:
        ProjectSettingsService.config_override["cli.log_config"] = log_config

    ctx.ensure_object(dict)
    ctx.obj["verbosity"] = verbose
    try:  # noqa: WPS229
        project = Project.find()
        setup_logging(project)

        readonly = ProjectSettingsService(project).get("project_readonly")
        if readonly:
            project.readonly = True
        if project.readonly:
            logger.debug("Project is read-only.")

        # detect active environment
        selected_environment = None
        if no_environment or (environment and environment.lower() == "null"):
            logger.info("No environment is active")
        elif environment:
            selected_environment = environment
        elif project.meltano.default_environment:
            selected_environment = project.meltano.default_environment
        # activate environment
        if selected_environment:
            project.activate_environment(selected_environment)
            logger.info(
                "Environment '%s' is active", selected_environment  # noqa: WPS323
            )

        ctx.obj["project"] = project
    except ProjectNotFound:
        ctx.obj["project"] = None
    except IncompatibleVersionError:
        click.secho(
            "This Meltano project is incompatible with this version of `meltano`.",
            fg="yellow",
        )
        click.echo(
            "For more details, visit http://meltano.com/docs/installation.html#upgrading-meltano-version"
        )
        sys.exit(3)
